#!/usr/bin/python3

import sys, os
import re

SOURCE_DIRS = ['reckless', 'examples', 'tests', 'performance_log']

ROOTDIR = os.path.relpath(os.path.join(os.path.dirname(__file__), '..'))

def main():
    read_license()
    for source_dir in SOURCE_DIRS:
        for dirpath, dirnames, filenames in os.walk(os.path.join(ROOTDIR, source_dir)):
            for filename in filenames:
                fullpath = os.path.join(dirpath, filename)
                fullpath = os.path.relpath(fullpath)
                if filename.endswith('.cpp') or filename.endswith('.hpp'):
                    update_license_header(fullpath)
    return 0

def read_license():
    global LICENSE

    license = os.path.join(ROOTDIR, 'scripts', 'license_header.txt')
    with open(license) as f:
        license = f.read()
    license = license.split('\n')
    if license[-1] == '':
        license = license[:-1]

    LICENSE = ['/* ' + license[0]] + [' * ' + x for x in license[1:]] + [' */']
    LICENSE = [x.rstrip() for x in LICENSE]
    LICENSE = '\n'.join(LICENSE) + '\n'

def update_license_header(path):
    with open(path) as f:
        original_data = f.read()
    data = original_data
    if data.startswith('/*'):
        i = data.find('*/\n')
        if i != -1:
            data = data[i+3:]
    data = LICENSE + data
    if original_data == data:
        return
    with open(path, 'w') as f:
        f.write(data)
        print(path)

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/python2
import sys
from sys import argv, stderr
from getopt import gnu_getopt
import os.path
from math import pi, sqrt, exp

ALL_LIBS = ['nop', 'reckless', 'stdio', 'fstream', 'boost_log', 'spdlog', 'g3log']
ALL_TESTS = ['periodic_calls', 'call_burst', 'write_files'] #, 'mandelbrot']

THREADED_TESTS = {'call_burst', 'mandelbrot'}

# color palette from colorbrewer2.org
COLORS = [
    '#1b9e77',
    '#d95f02',
    '#7570b3',
    '#e7298a',
    '#66a61e',
    '#e6ab02',
    '#a6761d',
]

def get_rdtsc_frequency():
    with open('/sys/devices/system/cpu/cpu0/cpufreq/scaling_max_freq', 'r') as f:
        return int(f.read())*1000

RDTSC_FREQUENCY = get_rdtsc_frequency()
TIME_SCALE = 1000000000.0/RDTSC_FREQUENCY

def get_default_window(test):
    table = {
        'periodic_calls': 8,
        'call_burst': 128,
        'write_files': 128,
        'mandelbrot': 1
    }

    return table.get(test, 1)

def pretty_name(name):
    name_table = {
            'nop': 'no operation',
            'stdio': 'fprintf (C)',
            'fstream': 'std::fstream (C++)',
            'reckless': 'reckless',
            'periodic_calls': 'periodic calls',
            'call_burst': 'single call burst',
            'write_files': 'heavy disk I/O',
            'mandelbrot': 'mandelbrot render'
    }

    return name_table.get(name, name)

def lib_color(name):
    color_table = {
            'nop': COLORS[0],
            'reckless': COLORS[3],
            'spdlog': COLORS[2],
            'stdio': COLORS[1],
            'fstream': COLORS[4],
            'boost_log': COLORS[5],
            'g3log': COLORS[6]
            }
    return color_table[name]


def average(average_window, data):
    window = []
    newdata = []
    for v in data:
        window.append(v)
        if len(window) == average_window:
            newdata.append(sum(window)/average_window)
            del window[0]
    return newdata

def average2(average_window, data):
    window = []
    newdata = []
    for v in data:
        window.append(v)
        if len(window) == average_window:
            newdata.extend([sum(window)/average_window]*average_window)
            del window[:]
    #if len(window) != 0:
    #    window.extend([0]*(average_window-len(window)))
    #    newdata.append(sum(window)/average_window)
    return newdata

def gaussian_kernel(width):
    kernel = []
    std = float(width)/6
    std = std*std
    C = 1.0/sqrt(2*pi*std)
    for x in range(0, width):
        xd = x - float(width)/2
        kernel.append(C*exp(-xd*xd/(2*std)))

    n = sum(kernel)
    kernel = [x/n for x in kernel]

    return kernel

def gaussian_average(average_window, data):
    kernel = gaussian_kernel(average_window)
    window = []
    newdata = []
    for v in data:
        window.append(v)
        if len(window) == average_window:
            newdata.append(sum([x*k for x, k in zip(window, kernel)]))
            del window[0]
    return newdata

def parse_ranges(s):
    ranges = s.split(',')
    result = []
    for r in ranges:
        parts = r.split('-')
        if len(parts)>2:
            raise ValueError("Invalid range specification: " + r)
        start=int(parts[0])
        if len(parts) == 2:
            end=int(parts[1])
        else:
            end = start
        start, end = min(start, end), max(start, end)
        result.extend(list(range(start, end+1)))
    return result

def main():
    opts, args = gnu_getopt(argv[1:], 'l:t:c:w:h', ['libs=', 'tests=',
        'threads=', 'window=', 'file=', 'top=', 'bottom=', 'iterations=',
        'title=', 'help'])
    libs = None
    tests = None
    threads = None
    window = None
    filename = None
    width = 858
    height = 858
    show_help = len(args) != 0
    top = None
    bottom = None
    iterations = None
    title = None

    for option, value in opts:
        if option in ('-l', '--libs'):
            libs = [lib.strip() for lib in value.split(',')]
        elif option in ('-t', '--tests'):
            tests = [test.strip() for test in value.split(',')]
        elif option in ('-c', '--threads'):
            threads = parse_ranges(value)
        elif option in ('-h', '--help'):
            show_help = True
        elif option in ('-w', '--window'):
            window = int(value)
        elif option == '--file':
            filename = value
        elif option == '--bottom':
            bottom = int(value)
        elif option == '--top':
            top = int(value)
        elif option == '--iterations':
            iterations = int(value)
        elif option == '--title':
            title = value

    if show_help:
        stderr.write(
            'usage: plot.py [OPTIONS]\n'
            'where OPTIONS are:\n'
            '-t,--tests    TESTS      comma-separated list of tests to plot\n'
            '-l,--libs     LIBS       comma-separated list of libs to plot\n'
            '-c,--threads  THREADS    thread-counts to include in a comma-separated list (e.g. 1-2,4)\n'
            '-w,--window   SIZE       Size of moving-average window\n'
            '--top         TOP        Top y coordinate for chart\n'
            '--bottom      BOTTOM     Bottom y coordinate for chart\n'
            '--iterations  ITERATIONS Number of iterations to include\n'
            '--title    TITLE      Plot title\n'
            '-h,--help        show this help\n'
            'Available libraries: {}\n'
            'Available tests: {}\n'.format(
                ','.join(ALL_LIBS), ','.join(ALL_TESTS)))
        return 1

    if libs is None:
        libs = sorted(ALL_LIBS)
    if tests is None:
        tests = sorted(ALL_TESTS)
    if threads is None:
        threads = list(range(1, 5))

    plot(libs, tests, threads, window, top, bottom, iterations, filename, width, height, title)
    return 0

def plot(libs, tests, threads_list, window, top, bottom, iterations, plot_filename, width, height, title, dpi=96):
    import matplotlib
    matplotlib.rc('font', size=10)
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots()

    def single_plot(filename, test, name, window, color=None):
        with open(filename, 'r') as f:
            lines = f.readlines()
        data = []
        for line in lines:
            assert line.endswith('\n')
            start, end = line[:-1].split(' ')
            start, end = int(start), int(end)
            data.append((end - start)*TIME_SCALE)
        if window is None:
            window = 1 #get_default_window(test)
        if window != 1:
            data = gaussian_average(window, data)
        if iterations is not None and iterations<len(data):
            data = data[:iterations]
        ax.plot(data, '-', label=name, color=color, linewidth=1)

    for test in tests:
        for lib in libs:
            color = lib_color(lib)
            base_name = []
            if len(libs)>1:
                base_name.append(pretty_name(lib))
            if len(tests)>1:
                base_name.append(pretty_name(test))
            if test in THREADED_TESTS:
                for threads in threads_list:
                    name = base_name[:]
                    filename = "results/%s-%s-%d.txt" % (test, lib, threads)
                    if len(threads_list)>1:
                        name.append("%d threads" % threads)
                    single_plot(filename, test, ', '.join(name), window, color)
            else:
                filename = "results/%s-%s.txt" % (test, lib)
                single_plot(filename, test, ', '.join(base_name), window, color)

    if top is not None:
        ax.set_ylim(ymax=top)
    if bottom is not None:
        ax.set_ylim(ymin=bottom)

    legend = ax.legend()
    # set the linewidth of each legend object
    for legobj in legend.legendHandles:
        legobj.set_linewidth(4)

    plt.xlabel('Iteration')
    plt.ylabel('Latency (nanoseconds)')
    if title is not None:
        fig.canvas.set_window_title(title)
    if plot_filename is None:
        plt.show()
    else:
        fig.set_size_inches(width/dpi, height/dpi)
        plt.savefig(plot_filename, dpi=dpi)

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/python3
from sys import stdout
import os, errno
from os.path import join
import shutil
import subprocess

for root, dirs, files in os.walk('.'):
    assert root.startswith('.')
    if root.startswith('./'):
        root = root[2:]
    else:
        root = root[1:]
    if '.gitignore' not in files:
        continue
    
    files_to_remove = []
    preserved_lines = []
    with open(join(root, '.gitignore'), 'rU') as f:
        live = False
        for line in f.readlines():
            if line == "##### TUP GITIGNORE #####\n":
                live = True
            if live:
                if line.startswith('/') and line != '/.gitignore\n':
                    files_to_remove.append(line[1:-1])
            else:
                preserved_lines.append(line)

    if len(files_to_remove) != 0:
        stdout.write(root + '/\n')
    for name in files_to_remove:
        stdout.write('  ')
        stdout.write(name + '\n')
        try:
            os.unlink(join(root, name))
        except OSError as e:
            if e.errno == errno.ENOENT:
                pass
            else:
                raise
        
    if len(preserved_lines) == 0:
        os.unlink(join(root, '.gitignore'))
    else:
        with open(join(root, '.gitignore'), 'w') as f:
            f.write(''.join(preserved_lines))

shutil.rmtree(".tup")
subprocess.check_call(['tup', 'init'])



#!/usr/bin/python3

import sys
from sys import argv, stderr, stdout
from getopt import gnu_getopt
import numpy as np

ALL_LIBS = ['nop', 'reckless', 'stdio', 'fstream', 'boost_log', 'spdlog', 'g3log']
ALL_TESTS = ['periodic_calls', 'call_burst', 'write_files', 'mandelbrot']
THREADED_TESTS = {'call_burst', 'mandelbrot'}

def main():
    opts, args = gnu_getopt(argv[1:], 'l:t:c:n:o:p:h', ['libs=', 'tests=', 'threads=', 'normalizer=', 'offset=', 'precision=', 'help'])
    libs = None
    tests = None
    threads = None
    normalizer = None
    offset = None
    precision = None
    show_help = len(args) != 0

    for option, value in opts:
        if option in ('-l', '--libs'):
            libs = [lib.strip() for lib in value.split(',')]
        elif option in ('-t', '--tests'):
            tests = [test.strip() for test in value.split(',')]
        elif option in ('-c', '--threads'):
            threads = parse_ranges(value)
        elif option in ('-n', '--normalizer'):
            normalizer = float(value)
        elif option in ('-o', '--offset'):
            offset = float(value)
        elif option in ('-p', '--precision'):
            precision = int(value)
        elif option in ('-h', '--help'):
            show_help = True

    if show_help:
        stderr.write(
            'usage: statistics.py [OPTIONS]\n'
            'where OPTIONS are:\n'
            '-t,--tests      TESTS      comma-separated list of tests to plot\n'
            '-l,--libs       LIBS       comma-separated list of libs to plot\n'
            '-c,--threads    THREADS    thread-counts to include in a comma-separated list (e.g. 1-2,4)\n'
            '-n,--normalizer NORMALIZER Normalizer value\n'
            '-o,--offset     OFFSET     Offset/displacement for values\n'
            '-p,--precision  PRECISION  Precision.\n'
            '-h,--help        show this help\n'
            'Available libraries: {}\n'
            'Available tests: {}\n'.format(
                ','.join(ALL_LIBS), ','.join(ALL_TESTS)))
        return 1

    if libs is None:
        libs = sorted(ALL_LIBS)
    if tests is None:
        tests = sorted(ALL_TESTS)
    if threads is None:
        threads = list(range(1, 5))

    make_stats(libs, tests, threads, normalizer, offset, precision)
    return 0

def parse_ranges(s):
    ranges = s.split(',')
    result = []
    for r in ranges:
        parts = r.split('-')
        if len(parts)>2:
            raise ValueError("Invalid range specification: " + r)
        start=int(parts[0])
        if len(parts) == 2:
            end=int(parts[1])
        else:
            end = start
        start, end = min(start, end), max(start, end)
        result.extend(list(range(start, end+1)))
    return result

def make_stats(libs, tests, threads_list, normalizer=None, offset=None, precision=None):
    def single_file_stats(filename, columns):
        with open(filename, 'r') as f:
            lines = f.readlines()
        data = []
        for line in lines:
            assert line.endswith('\n')
            numbers = line[:-1].split(' ')
            numbers = [int(n) for n in numbers]
            if len(numbers) == 1:
                data.append(numbers[0])
            else:
                data.append(numbers[1] - numbers[0])
        if offset is not None:
            data = [x - offset for x in data]
        data = np.array(data)
        mean = np.mean(data)
        if normalizer is not None:
            mean = mean/normalizer
            data = [x/normalizer for x in data]
        low, high = np.percentile(data, [25, 75])
        mad = np.mean(np.absolute(data - mean))
        std = np.std(data)
        cols = [mean, high - low, mad, std]
        if precision is None:
            cols = ["%g" % x for x in cols]
        else:
            fmt = "%%.%df" % precision
            cols = [fmt % x for x in cols]
        columns.extend(cols)
        return mean

    rows = []
    for test in tests:
        for lib in libs:
            columns = []
            if len(lib) > 1:
                columns.append(lib)
            base_name = []
            if test in THREADED_TESTS:
                for threads in threads_list:
                    name = base_name[:]
                    filename = "results/%s-%s-%d.txt" % (test, lib, threads)
                    mean = single_file_stats(filename, columns)
            else:
                filename = "results/%s-%s.txt" % (test, lib)
                mean = single_file_stats(filename, columns)
            rows.append((mean, columns))

    rows = [r for _, r in sorted(rows)]
    rows.insert(0, ["Library", "Ticks", "IQR", "MAD", "Std deviation"])
    if normalizer is not None:
        rows[0][1] = "Relative time"

    colwidths = [0]*len(rows[0])
    for row in rows:
        widths = [len(x) for x in row]
        colwidths = [max(a, b) for a,b in zip(colwidths, widths)]
    first_row = True
    for row in rows:
        first_col = True
        for col, width in zip(row, colwidths):
            if not first_col:
                stdout.write(' | ')
            first_col = False
            stdout.write(' '*(width - len(col)))
            stdout.write(col)
        stdout.write('\n')
        if first_row:
            line = ['-'*w for w in colwidths]
            line = '-|-'.join(line) + '\n'
            stdout.write(line)
            first_row = False

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/python3

import sys, os, time, subprocess, errno
from sys import stdout, stderr, argv
from getopt import gnu_getopt

ALL_LIBS = ['nop', 'reckless', 'stdio', 'fstream', 'boost_log', 'spdlog', 'g3log']
ALL_TESTS = ['periodic_calls', 'call_burst', 'write_files', 'mandelbrot']

SINGLE_SAMPLE_TESTS = {'mandelbrot'}
THREADED_TESTS = {'call_burst', 'mandelbrot'}
TESTS_WITH_DRY_RUN = {'call_burst', 'periodic_calls'}
MAX_THREADS = 8

# /run_benchmark.py -t mandelbrot  1448.92s user 64.87s system 208% cpu 12:04.87 total
# with SINGLE_SAMPLE_TEST_ITERATIONS=2 means we should have
# about 80 for 8 hours runtime
#
# time ./run_benchmark.py -t mandelbrot
# mandelbrot: fstream/1234 nop/1234 pantheios/1234 reckless/1234 spdlog/1234 stdio/1234
# ./run_benchmark.py -t mandelbrot  58192.78s user 2484.45s system 208% cpu 8:04:15.16 total
#
# Then 100 = 10 hours runtime
SINGLE_SAMPLE_TEST_ITERATIONS = 100

def main():
    opts, args = gnu_getopt(argv[1:], 'l:t:h', ['libs', 'tests', 'help'])
    libs = None
    tests = None
    show_help = len(args) != 0

    for option, value in opts:
        if option in ('-l', '--libs'):
            libs = [lib.strip() for lib in value.split(',')]
        elif option in ('-t', '--tests'):
            tests = [test.strip() for test in value.split(',')]
        elif option in ('-h', '--help'):
            show_help = True

    if show_help:
        stderr.write(
            'usage: run_benchmark.py [OPTIONS]\n'
            'where OPTIONS are:\n'
            '-t,--tests TESTS comma-separated list of tests to run\n'
            '-l,--libs  LIBS  comma-separated list of libs to benchmark\n'
            '-h,--help        show this help\n'
            'Available libraries: {}\n'
            'Available tests: {}\n'.format(
                ','.join(ALL_LIBS), ','.join(ALL_TESTS)))
        return 1

    if libs is None:
        libs = sorted(ALL_LIBS)
    if tests is None:
        tests = sorted(ALL_TESTS)

    run_tests(libs, tests)
    return 0

def run_tests(libs, tests):
    for test in tests:
        stdout.write(test + ':')
        stdout.flush()
        for lib in libs:
            stdout.write(' ' + lib)
            if test in THREADED_TESTS:
                stdout.write('/')
                for threads in range(1, MAX_THREADS+1):
                    stdout.write(str(threads))
                    stdout.flush()
                    success = run_test(lib, test, threads)
            else:
                stdout.flush()
                success = run_test(lib, test)

            if not success:
                stdout.write('[F]')

        stdout.write('\n')
        stdout.flush()

def reset():
    if os.path.exists('log.txt'):
        os.unlink('log.txt')
    for name in os.listdir('data'):
        os.unlink(os.path.join('data', name))
    subprocess.call('sync')

def run_test(lib, test, threads = None):
    binary_name = test + '-' + lib
    if threads is not None:
        binary_name += '-' + str(threads)

    def run(out_file):
        try:
            p = subprocess.Popen([binary_name], executable='./' + binary_name, stdout=out)
            p.wait()
        except OSError as e:
            if e.errno == errno.ENOENT:
                return False
            else:
                raise

    if test in TESTS_WITH_DRY_RUN:
        with open(os.devnull, 'w') as out:
            run(out)
    else:
        busy_wait(0.5)

    txt_name = binary_name + '.txt'
    with open('results/' + txt_name, 'w') as out:
        total_iterations = 1
        if test in SINGLE_SAMPLE_TESTS:
            total_iterations = SINGLE_SAMPLE_TEST_ITERATIONS
        for i in range(0, total_iterations):
            reset()
            run(out)
    return True

def busy_wait(period):
    end = time.time() + period
    while time.time() < end:
        pass

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, shutil
from conans import ConanFile, tools, AutoToolsBuildEnvironment, MSBuild
from conans.errors import ConanException

MSDEV_PLATFORM_SHORT_NAMES = {
    'x86': 'x86',
    'x86_64': 'x64'
}

class RecklessConan(ConanFile):
    name = 'reckless'
    license = 'MIT'
    version = '3.0.1'
    url = 'https://github.com/mattiasflodin/reckless'
    description = """Reckless is an extremely low-latency, high-throughput logging library."""
    settings = 'arch', 'compiler', 'build_type'
    exports_sources = (
        "reckless/src/*",
        "reckless/include/*",
        "Makefile.conan",
        "reckless.sln",
        "common.props",
        "*.vcxproj"
    )

    def configure(self):
        if self._gcc_compatible() and self.settings.compiler.libcxx == "libstdc++":
            raise ConanException("This package is only compatible with libstdc++11. "
                "Please run with -s compiler.libcxx=libstdc++11.")

    def build(self):
        if self.settings.compiler == "Visual Studio":
            env = MSBuild(self)
            env.build('reckless.sln', use_env=False, targets=['reckless:Rebuild'])
        else:
            env = AutoToolsBuildEnvironment(self)
            env.make(args=['-f', 'Makefile.conan'], target="clean")
            # Why doesn't Conan set CXX?
            vars = env.vars
            vars['CXX'] = str(self.settings.compiler)
            env.make(args=['-f', 'Makefile.conan'], vars=vars)

    def package(self):
        if self._gcc_compatible():
            self.copy('*.a', src='reckless/lib', dst='lib', keep_path=False)
        else:
            platform_short_name = MSDEV_PLATFORM_SHORT_NAMES[str(self.settings.arch)]
            configuration_name = str(self.settings.build_type).lower()
            build_directory = os.path.join('build', '%s-%s' % (platform_short_name, configuration_name))
            self.copy('*.lib', src=build_directory, dst='lib', keep_path=False)
            self.copy('*.pdb', src=build_directory, dst='lib', keep_path=False)

        self.copy('*.hpp', src='reckless/include', dst='include', keep_path=True)
        self.copy('*.cpp', src='reckless/src', dst='src', keep_path=True)
        self.copy('LICENSE.txt', src='src')

    def package_info(self):
        self.cpp_info.libs = ["reckless"]

    def _gcc_compatible(self):
        return str(self.settings.compiler) in ['gcc', 'clang', 'apple-clang']



#!/usr/bin/python2
from __future__ import print_function
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from sys import argv
import os.path

matplotlib.rc('font', size=10)

# color palette from colorbrewer2.org
COLORS = [
    '#a6cee3',
    '#1f78b4',
    '#b2df8a',
    '#33a02c',
    '#fb9a99',
    '#e31a1c',
    '#fdbf6f'
    ]
#COLORS = [
#    '#8dd3c7',
#    '#fb8072',
#    '#80b1d3',
#    #'#ffffb3', # this is too bright for white background
#    '#bebada',
#    '#fdb462',
#    '#b3de69',
#    '#fccde5'
#]

filename = None
if len(argv) > 1:
    filename = argv[1]
    width = 858
    height = 858

fig, ax = plt.subplots()

ONLY_OVERHEAD = 1
SCALE = 1000000.0/(1024*1024)
#LOG_LINES = 1048576
MAX_CORES = 8
LIBS = ['nop', 'reckless', 'spdlog', 'g3log', 'stdio', 'fstream', 'boost_log']
#LIBS = ['nop', 'reckless', 'stdio', 'fstream']
if ONLY_OVERHEAD:
    del COLORS[-1]
    del LIBS[0]
ind = np.arange(MAX_CORES)
WIDTH=1.0/(len(LIBS)+2)
GROUP_WIDTH = WIDTH*len(LIBS)
GROUP_OFFSET = (1.0 - GROUP_WIDTH)/2

def read_timing(lib, cores, offset=0.0):
    with open(os.path.join('results', 'mandelbrot-%s-%d.txt' % (lib, cores)), 'r') as f:
        data = f.readlines()
    data = sorted([float(x)/1000.0+offset for x in data])
    # Use interquartile range as measure of scale.
    low, high = np.percentile(data, [25, 75])

    # Samples generally center around a minimum point that represents the ideal
    # running time. There are no outliers below the ideal time; execution does
    # not accidentally take some kind of short cut. At least, not for this
    # particular benchmark. It could be conceivable for other tests where we
    # are I/O-bound and we get lucky with the cache, but this test is not
    # I/O-bound.
    #
    # However, outliers do exist on the other end of the spectrum. Something
    # completely unrelated sometimes happens in the system and takes tons of
    # CPU cycles from our benchmark. We don't want to include these because
    # they make it more difficult to reproduce the same statistics on multiple
    # runs. Or, in statistics speak, they make our benchmark less robust and
    # less efficient. To counter this we get rid of the worst outliers on the
    # upper end, but we leave the lower end intact.
    outlier_index = int(len(data)*0.8)
    data = data[:outlier_index]
    mean = np.mean(data)
    return low, mean, high

offsets = []
for cores in range(1, MAX_CORES+1):
    if ONLY_OVERHEAD:
        _, mean, _ = read_timing('nop', cores)
        offsets.append(-mean)
    else:
        offsets.append(0.0)

rects = []
for index, lib in enumerate(LIBS):
    means = []
    error = [[], []]

    for cores in range(1, MAX_CORES+1):
        low, mean, high = read_timing(lib, cores, offsets[cores-1])
        low, mean, high = low*SCALE, mean*SCALE, high*SCALE
        print(lib, cores, mean, low, high)
        means.append(mean)
        error[0].append(mean - low)
        error[1].append(high - mean)
    if not ONLY_OVERHEAD:
        # Standard deviation is so small that it just clutters the plot in this
        # view
        error = None
    rects.append(ax.bar(ind+index*WIDTH + GROUP_OFFSET, means, WIDTH,
        color=COLORS[index], yerr=error, ecolor='black'))

ax.legend( rects, LIBS )

if ONLY_OVERHEAD:
    ax.set_ylabel('Average call latency (microseconds)')
else:
    ax.set_ylabel('Total seconds')
ax.set_xlabel('Number of worker threads')
ax.set_xticks(ind + GROUP_OFFSET + GROUP_WIDTH/2)
ax.set_xticklabels([str(x) for x in range(1, MAX_CORES+1)])
#ax.set_title('1024x1024 mandelbrot set render, one log line per pixel (i.e. 1.05 million log lines)')

if filename is None:
    plt.show()
else:
    dpi = 96
    fig.set_size_inches(width/dpi, height/dpi)
    plt.savefig(filename, dpi=dpi)
